package systems.kumi.expalert

import android.content.Context
import android.content.SharedPreferences
import android.preference.PreferenceManager
import okhttp3.MediaType.Companion.toMediaTypeOrNull
import okhttp3.OkHttpClient
import okhttp3.RequestBody.Companion.toRequestBody
import java.security.KeyStore
import java.security.cert.CertificateFactory
import javax.net.ssl.TrustManager
import javax.net.ssl.TrustManagerFactory
import javax.net.ssl.X509TrustManager

class ServerConnection(context: Context) {
    private val context: Context = context
    private val preferences: SharedPreferences = PreferenceManager.getDefaultSharedPreferences(context)
    private val client = this.getOkHttpClient()

    private fun getOkHttpClient(): OkHttpClient {
        val cf = CertificateFactory.getInstance("X.509")
        val caInput = context.resources.openRawResource(R.raw.jukumi)
        val ca = cf.generateCertificate(caInput)
        caInput.close()

        val keyStoreType = KeyStore.getDefaultType()
        val keyStore = KeyStore.getInstance(keyStoreType)
        keyStore.load(null, null)
        keyStore.setCertificateEntry("jukumi", ca)

        val tmfAlgorithm = TrustManagerFactory.getDefaultAlgorithm()
        val tmf = TrustManagerFactory.getInstance(tmfAlgorithm)
        tmf.init(keyStore)

        val sslContext = javax.net.ssl.SSLContext.getInstance("TLS")

        val defaultTrustManagerFactory = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm())
        defaultTrustManagerFactory.init(null as KeyStore?)
        val defaultTrustManagers = defaultTrustManagerFactory.trustManagers

        val combinedTrustManagers = mutableListOf<TrustManager>()

        defaultTrustManagers.forEach { defaultTrustManager ->
            if (defaultTrustManager is X509TrustManager) {
                combinedTrustManagers.add(defaultTrustManager)
            }
        }

        combinedTrustManagers.addAll(tmf.trustManagers)

        sslContext.init(null, arrayOf(MergeTrustManager(combinedTrustManagers)), null)

        val x509TrustManager = combinedTrustManagers.first() as X509TrustManager

        return OkHttpClient.Builder()
            .sslSocketFactory(sslContext.socketFactory, x509TrustManager)
            .build()
    }

    private fun getSetting(key: String, default: String = ""): String? {
        var value = preferences.getString(key, default)
        return if (value.isNullOrBlank()) default else value
    }

    private fun getSetting(key: String): String? {
        return this.getSetting(key, "")
    }

    private fun getHeartbeatURL(): String {
        val hostname = this.getSetting("hostname")
        return "https://$hostname/apps/heartbeat"
    }

    private fun getRegisterURL(): String {
        val hostname = this.getSetting("hostname")
        return "https://$hostname/apps/register"
    }

    private fun getSolveURL(): String {
        val hostname = this.getSetting("hostname")
        return "https://$hostname/apps/solve"
    }

    private fun getPanicURL(): String {
        val hostname = this.getSetting("hostname")
        return "https://$hostname/apps/panic"
    }

    private fun getUpdateURL(): String {
        val hostname = this.getSetting("hostname")
        return "https://$hostname/apps/update"
    }

    private fun getAppKey(): String? {
        return this.getSetting("key")
    }

    fun heartbeat(): String? {
        val url = this.getHeartbeatURL()
        val request = client.newBuilder().build().newCall(
            okhttp3.Request.Builder()
                .url(url)
                .header("User-Agent", "EXPAlert Android")
                .post(
                    "{\"key\": \"${this.getAppKey()}\"}"
                        .toRequestBody("application/json".toMediaTypeOrNull()),
                )
                .build()
        ).execute()
        return request.body?.string()
    }

    fun register(message: String): String? {
        val url = this.getRegisterURL()
        val request = client.newBuilder().build().newCall(
            okhttp3.Request.Builder()
                .url(url)
                .header("User-Agent", "EXPAlert Android")
                .post(message.toRequestBody("application/json".toMediaTypeOrNull()))
                .build()
        ).execute()
        return request.body?.string()
    }

    fun solve(message: String): String? {
        val url = this.getSolveURL()
        val request = client.newBuilder().build().newCall(
            okhttp3.Request.Builder()
                .url(url)
                .header("User-Agent", "EXPAlert Android")
                .post(message.toRequestBody("application/json".toMediaTypeOrNull()))
                .build()
        ).execute()
        return request.body?.string()
    }

    fun panic(message: String): String? {
        val url = this.getPanicURL()
        val request = client.newBuilder().build().newCall(
            okhttp3.Request.Builder()
                .url(url)
                .header("User-Agent", "EXPAlert Android")
                .post(message.toRequestBody("application/json".toMediaTypeOrNull()))
                .build()
        ).execute()
        return request.body?.string()
    }

    fun update(): String? {
        val url = this.getUpdateURL()
        val request = client.newBuilder().build().newCall(
            okhttp3.Request.Builder()
                .url(url)
                .header("User-Agent", "EXPAlert Android")
                .get()
                .build()
        ).execute()
        return request.body?.string()
    }

}